/*
** File: bitstrip.c - strip header from a .bit file and save only the SelectMap
**                    data.
**
** (C) Copyright Alpha Data 2012
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _stprintf_s snprintf
# endif

#else

/* Linux or VxWorks */
# define _T(x) x
# define _stprintf_s snprintf
# define _tmain main
# define _tprintf printf
# define _tcserror strerror
# define _tfopen fopen
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <admxrc3.h>

#include <sdk_common.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#if !defined(_MSC_VER) || !defined(WIN32)
static int
_tfopen_s(
  FILE** ppFile,
  const TCHAR* pFilename,
  const TCHAR* pMode)
{
  FILE* pFile;

  pFile = _tfopen(pFilename, pMode);
  if (NULL != pFile) {
    *ppFile = pFile;
    return 0;
  } else {
    return errno;
  }
}

static int
_tcserror_s(
  TCHAR* pBuffer,
  size_t n,
  int err)
{
  const TCHAR* p = _tcserror(err);
  size_t i;

  if (n == 0) {
    return 0;
  }
  if (NULL == pBuffer) {
    return EINVAL;
  }
  for (i = 0; i < n; i++) {
    pBuffer[i] = p[i];
    if (p[i] == _T('\0')) {
      break;
    }
  }
  pBuffer[n - 1] = _T('\0');
  return 0;
}
# endif

int
_tmain(
  int argc,
  TCHAR* argv[])
{
  int ret = 0;
  ADMXRC3_STATUS status;
  ADMXRC3_BITSTREAM* pBitstream = NULL;
  Option options[1];
  OptionType paramType[1];
  TCHAR* paramInfo[1];
  TCHAR* paramMeaning[1];
  Arguments arguments = g_initArguments;
  TCHAR* pInputName;
  bool_t bWriteOutput;
  TCHAR* pOutputName;
  FILE* pOutFile = NULL;

  options[0].key = _T("o");
  options[0].type = OptionPath;
  options[0].def.pPathVal = _T("");
  options[0].meaning = _T("Output filename");
  paramType[0] = OptionPath;
  paramInfo[0] = _T("<input.bit>");
  paramMeaning[0] = _T(".bit file to be stripped");
  arguments.nOption = ARRAY_LENGTH(options);
  arguments.option = options;
  arguments.nParamType = ARRAY_LENGTH(paramType);
  arguments.minNParam = ARRAY_LENGTH(paramType);
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;
  arguments.bNoStdOptions = TRUE;

  if (sampleParseCommandLine(_T("bitstrip"), argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }
  if (arguments.nParam < 1) {
    _tprintf(_T("*** .bit filename must be specified.\n"));
    ret = -1;
    goto done;
  }
  pInputName = arguments.param[0].pPathVal;
  bWriteOutput = options[0].specified;
  pOutputName = options[0].value.pPathVal;

  status = ADMXRC3_LoadBitstream(pInputName, &pBitstream);
  if (ADMXRC3_SUCCESS != status) {
    _tprintf(_T("*** Failed to load '%s': %s\n"), pInputName, ADMXRC3_GetStatusString(status, TRUE));
    ret = -2;
    goto done;
  }
  _tprintf(_T("'%s' -> device %s, SelectMap data length %llu(0x%llx) bytes.\n"),
    pInputName, pBitstream->Identifier, (unsigned long long)pBitstream->Length, (unsigned long long)pBitstream->Length);

  if (bWriteOutput) {
    size_t nWritten;
    int err;
    TCHAR buf[128];

    err = _tfopen_s(&pOutFile, pOutputName, _T("wb"));
    if (0 != err) {
      buf[0] = _T('\0');
      _tcserror_s(buf, ARRAY_LENGTH(buf), err);
      _tprintf(_T("*** Failed to open '%s' for writing: %s\n"), pOutputName, buf);
      ret = -3;
      goto done;
    }
    nWritten = fwrite(&pBitstream->Data[0], sizeof(uint8_t), pBitstream->Length, pOutFile);
    if (nWritten != pBitstream->Length) {
      _tprintf(_T("*** Only managed to write %llu(0x%llx) of %llu(0x%llx) bytes for file '%s'\n"),
        (unsigned long long)nWritten, (unsigned long long)nWritten,
        (unsigned long long)pBitstream->Length, (unsigned long long)pBitstream->Length,
        pOutputName);
      ret = -3;
      goto done;
    }
    fclose(pOutFile);
    pOutFile = NULL;
    _tprintf(_T("Wrote %llu(0x%llx) bytes to '%s'.\n"),
      (unsigned long long)nWritten, (unsigned long long)nWritten, pOutputName);
  }

done:
  if (NULL != pOutFile) {
    fclose(pOutFile);
    pOutFile = NULL;
  }
  if (NULL != pBitstream) {
    ADMXRC3_UnloadBitstream(pBitstream);
    pBitstream = NULL;
  }
  sampleCleanupCommandLine(&arguments);
  return ret;
}
