/*
** itest.c - Linux-specific example demonstrating interrupt wait
**           functionality in the ADMXRC3 API.
**
** (C) Copyright 2010-2011 Alpha Data
**
** Demonstrates how to wait for FPGA interrupt notifications using the non-
** blocking direct call technique in an interrupt thread. Also demonstrates
** how to make the interrupt thread terminate cleanly when the application
** shuts down.
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <pthread.h>
#include <semaphore.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>

/*
** FPGA register addresses for the ITest design, in OCP Direct Slave address
** space.
*/
#define REGADDR_ITEST_TEST   (0x000U) /* Test register; write 1 to a bit to generate an interrupt */
#define REGADDR_ITEST_ENABLE (0x180U) /* Enable register in interrupt controller */
#define REGADDR_ITEST_CLEAR  (0x184U) /* Write 1 to a bit to clear corresponding STATUS bit */
#define REGADDR_ITEST_STATUS (0x18CU) /* Status register in interrupt controller */

/*
** FPGA register addresses for the UBER design, in OCP Direct Slave address
** space.
*/
#define REGADDR_UBER_TEST    (0xC0U) /* Test register, write 1 to a bit to generate an interrupt */
#define REGADDR_UBER_ISTAT   (0xC4U) /* Status register */
#define REGADDR_UBER_IMASK   (0xC8U) /* Mask register */
#define REGADDR_UBER_IARM    (0xCCU) /* Rearm register */

typedef struct _ThreadContext {
  ADMXRC3_HANDLE hCard;
  unsigned long signalledCount;
  unsigned int targetIndex;
  unsigned int window;
  sem_t waitingSemaphore;
} ThreadContext;

static void*
interruptThreadFunctionITest(
  void* pArgument)
{
  ThreadContext* pContext = (ThreadContext*)pArgument;
  ADMXRC3_STATUS status;
  ADMXRC3_TICKET ticket;
  uint32_t istat;
  uint32_t dummy32 = 0; /* value doesn't matter, but avoid compiler whine */
  
  printf("Interrupt thread (ITest) started\n");
	
  ADMXRC3_InitializeTicket(&ticket);

  while (1) {
    /* Start a non-blocking wait for an FPGA interrupt */
    status = ADMXRC3_StartNotificationWait(pContext->hCard, &ticket, ADMXRC3_EVENT_FPGAINTERRUPT(pContext->targetIndex));
    if (status != ADMXRC3_PENDING) {
      printf("*** ADMXRC3_StartNotificationWait returned non-pending status: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      break;
    }

    /* We are now waiting */
    sem_post(&pContext->waitingSemaphore);

    /* Rearm the 'finti_l' line driven by the FPGA by forcing it high for 1 cycle */
    dummy32 = 0;
    status = ADMXRC3_Write(pContext->hCard, pContext->window, 0, REGADDR_ITEST_CLEAR, sizeof(dummy32), &dummy32);

    /* Finish the non-blocking wait for an FPGA interrupt */
    status = ADMXRC3_FinishNotificationWait(pContext->hCard, &ticket, TRUE);
    if (status != ADMXRC3_SUCCESS) {
      if (status == ADMXRC3_CANCELLED) {
        /* The main thread wants us to terminate */
        printf("Interrupt thread wait cancelled.\n");
        break;
      } else {
        printf("*** ADMXRC3_FinishNotificationWait failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
        break;
      }
    } else {
      pContext->signalledCount++;
    }

    /* We are no longer waiting */
    if (sem_trywait(&pContext->waitingSemaphore)) {
      /* The main thread wants us to terminate */
      printf("Interrupt thread sem_trywait failed.\n");
      break;
    }

    /* Get the pending FPGA interrupts */
    ADMXRC3_Read(pContext->hCard,	pContext->window, 0, REGADDR_ITEST_STATUS, sizeof(istat), &istat);

    /* Clear the pending FPGA interrupts */
    ADMXRC3_Write(pContext->hCard, pContext->window, 0, REGADDR_ITEST_CLEAR, sizeof(istat), &istat);
  }
  
  printf("Interrupt thread exiting.\n");

  return 0;
}

static void*
interruptThreadFunctionUber(
  void* pArgument)
{
  ThreadContext* pContext = (ThreadContext*)pArgument;
  ADMXRC3_STATUS status;
  ADMXRC3_TICKET ticket;
  uint32_t istat;
  uint32_t dummy32 = 0; /* value doesn't matter, but avoid compiler whine */
  
  printf("Interrupt thread (UBER) started\n");
	
  ADMXRC3_InitializeTicket(&ticket);

  while (1) {
    /* Start a non-blocking wait for an FPGA interrupt */
    status = ADMXRC3_StartNotificationWait(pContext->hCard, &ticket, ADMXRC3_EVENT_FPGAINTERRUPT(pContext->targetIndex));
    if (status != ADMXRC3_PENDING) {
      printf("*** ADMXRC3_StartNotificationWait returned non-pending status: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      break;
    }

    /* We are now waiting */
    sem_post(&pContext->waitingSemaphore);

    /* Rearm the 'FINTI_L' line driven by the FPGA by forcing it high for 1 cycle */
    dummy32 = 0;
    status = ADMXRC3_Write(pContext->hCard, pContext->window, 0, REGADDR_UBER_IARM, sizeof(dummy32), &dummy32);

    /* Finish the non-blocking wait for an FPGA interrupt */
    status = ADMXRC3_FinishNotificationWait(pContext->hCard, &ticket, TRUE);
    if (status != ADMXRC3_SUCCESS) {
      if (status == ADMXRC3_CANCELLED) {
        /* The main thread wants us to terminate */
        printf("Interrupt thread wait cancelled.\n");
        break;
      } else {
        printf("*** ADMXRC3_FinishNotificationWait failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
        break;
      }
    } else {
      pContext->signalledCount++;
    }

    /* We are no longer waiting */
    if (sem_trywait(&pContext->waitingSemaphore)) {
      /* The main thread wants us to terminate */
      printf("Interrupt thread sem_trywait failed.\n");
      break;
    }

    /* Get the pending FPGA interrupts */
    ADMXRC3_Read(pContext->hCard,	pContext->window, 0, REGADDR_UBER_ISTAT, sizeof(istat), &istat);

    /* Clear the pending FPGA interrupts */
    ADMXRC3_Write(pContext->hCard, pContext->window, 0, REGADDR_UBER_ISTAT, sizeof(istat), &istat);
  }
  
  printf("Interrupt thread exiting.\n");

  return 0;
}

int
main(
  int argc,
  char *argv[])
{
  const double clockFreq = 33.0e6;
  int ret = 0;
  unsigned int index, serial;
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFO cardInfo;
  unsigned long numInterruptGenerated;
  const char* pFilename;
  Arguments arguments = g_initArguments;
  Option options[3];
  ThreadContext threadContext = {
    ADMXRC3_HANDLE_INVALID_VALUE, /* hCard */
    0, /* signalledCount */
    0 /* targetIndex */,
    0 /* window */
  };
  pthread_t interruptThread;
  bool_t bInterruptThreadRunning = FALSE;
  bool_t bSemValid = FALSE, bUseUber, bShouldConfigure;
	uint32_t val32;
  int ch;

  options[0].key = "index";
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = "index of card to use";
  options[1].key = "sn";
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = "serial number of card to use";
  options[2].key = "uber";
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = "use uber design instead of itest";
	arguments.nOption = 3;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  if (sampleParseCommandLine("itest", argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bUseUber = options[2].value.booleanVal;  

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &threadContext.hCard);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to open card with index %lu: %s\n", (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(threadContext.hCard, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to get card info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }
	
  if (cardInfo.Model < 0x100) {
    /* Set local bus clock to nominal frequency */
    status = ADMXRC3_SetClockFrequency(threadContext.hCard, ADMXRC3_CLOCK_LCLK, 0, clockFreq, NULL);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to set LCLK to %.1fMHz: %s\n", clockFreq / 1.0e6f, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  }

  if (bUseUber) {
    pFilename = sampleGetFpgaFilename(threadContext.hCard, 0, "uber", "uber", &arguments, &bShouldConfigure);
  } else {
    pFilename = sampleGetFpgaFilename(threadContext.hCard, 0, "itest", "itest", &arguments, &bShouldConfigure);
  }
  if (NULL == pFilename) {
    printf("*** Failed to determine bitstream filename.\n");
    ret = -1;
    goto done;
  }

  if (bShouldConfigure) {
    status = ADMXRC3_ConfigureFromFile(threadContext.hCard, 0, 0, pFilename);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to load the bitstream '%s': %s\n", pFilename, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  } else {
    /* Running on a model such as the ADM-XRC-6T-ADV8, which doesn't have an FPGA that is configurable via the ADMXRC3 API */
  }

  /* Start up the thread that waits for interrupts. */
  threadContext.signalledCount = 0;
  threadContext.targetIndex = 0;
  threadContext.window = 0;
  sem_init(&threadContext.waitingSemaphore, TRUE, 0);
  bSemValid = TRUE;
  if (pthread_create(&interruptThread, NULL, bUseUber ? interruptThreadFunctionUber : interruptThreadFunctionITest, (void*)&threadContext)) {
    printf("*** Failed to start interrupt thread\n");
    ret = -1;
    goto done;
  }
  bInterruptThreadRunning = TRUE;
  
  if (bUseUber) {
  	/* Unmask all interrupts in the FPGA (UBER) */
	  val32 = cpu_to_le32(0x0U);
	  status = ADMXRC3_Write(threadContext.hCard, 0, 0, REGADDR_UBER_IMASK, sizeof(val32), &val32);
	  if (status != ADMXRC3_SUCCESS) {
		  printf("*** Write to UBER IMASK register failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
		  ret = -1;
		  goto done;
	  }
  } else {
  	/* Enable all interrupts in the FPGA (ITest) */
	  val32 = cpu_to_le32(0xFFFFFFFFU);
	  status = ADMXRC3_Write(threadContext.hCard, 0, 0, REGADDR_ITEST_ENABLE, sizeof(val32), &val32);
	  if (status != ADMXRC3_SUCCESS) {
		  printf("*** Write to ITest ENABLE register failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
		  ret = -1;
		  goto done;
	  }
  }
  
  /*
  ** Allow the user to generate interrupts by pressing the enter key.
  ** Ctrl-Z or entering 'q' / 'Q' exits this loop.
  */
  numInterruptGenerated = 0;
  do {
    printf("Enter 'q' to quit, or anything else to generate an interrupt:\n");
    ch = getchar();
    if (EOF == ch || 'q' == tolower(ch)) {
      ch = EOF;
    } else {
      val32 = cpu_to_le32(0x1U);
      status = ADMXRC3_Write(threadContext.hCard, threadContext.window, 0, bUseUber ? REGADDR_UBER_TEST : REGADDR_ITEST_TEST, sizeof(val32), &val32);
      if (status != ADMXRC3_SUCCESS) {
        printf("*** Write to FPGA_TEST register failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }
      
      numInterruptGenerated++;
    }
  } while (EOF != ch);

  printf("Generated %lu interrupts\n", numInterruptGenerated);
  printf("Interrupt thread saw %lu interrupt(s)\n", threadContext.signalledCount);

done:
  /* Make the interrupt thread to terminate cleanly, if it is running */
  if (bInterruptThreadRunning) {
    /* Wait until the interrupt thread is waiting */
    sem_wait(&threadContext.waitingSemaphore);
    /* Cancel any ongoing operations (including the interrupt thread's notification wait) */
    ADMXRC3_Cancel(threadContext.hCard);
    /* Join, so that the interrupt thread is cleaned up */
    if (pthread_join(interruptThread, NULL)) {
      printf("+++ Failed to join with interrupt thread\n");
    }
  }

  if (bSemValid) {
    sem_destroy(&threadContext.waitingSemaphore);
    bSemValid = FALSE;
  }

  if (threadContext.hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(threadContext.hCard);
    threadContext.hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
