/*
** itest.c - example demonstrating interrupt wait functionality in
**           the ADMXRC3 API via Win32 events.
**
** (C) Copyright 2009-2011 Alpha Data
*/

#include <windows.h>
#include <stdio.h>
#include <ctype.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>

/*
** FPGA register addresses for the ITest design, in OCP Direct Slave address
** space.
*/
#define REGADDR_ITEST_TEST   (0x000U) /* Test register; write 1 to a bit to generate an interrupt */
#define REGADDR_ITEST_ENABLE (0x180U) /* Enable register in interrupt controller */
#define REGADDR_ITEST_CLEAR  (0x184U) /* Write 1 to a bit to clear corresponding STATUS bit */
#define REGADDR_ITEST_STATUS (0x18CU) /* Status register in interrupt controller */

/*
** FPGA register addresses for the UBER design, in OCP Direct Slave address
** space.
*/
#define REGADDR_UBER_TEST    (0xC0U) /* Test register, write 1 to a bit to generate an interrupt */
#define REGADDR_UBER_ISTAT   (0xC4U) /* Status register */
#define REGADDR_UBER_IMASK   (0xC8U) /* Mask register */
#define REGADDR_UBER_IARM    (0xCCU) /* Rearm register */

typedef struct _ThreadContext {
  ADMXRC3_HANDLE hCard;
  HANDLE hEvent;
  BOOL bTerminate;
  unsigned long signalled;
} ThreadContext;

static DWORD WINAPI
interruptThreadITest(
  void* pArgument)
{
  ThreadContext* pContext = (ThreadContext*)pArgument;
	DWORD result;
	DWORD istat;

	_tprintf(_T("Interrupt thread (ITest) started\n"));
	
	while (!pContext->bTerminate) {
		result = WaitForSingleObject(pContext->hEvent, INFINITE);
		if (result != WAIT_OBJECT_0) {
      DWORD lastError = GetLastError();

			_tprintf(_T("*** Interrupt thread failed to wait for event (Win32 error %lu(0x%lX))\n"),
        (unsigned long)lastError, (unsigned long)lastError);
			break;
		}

		if (!pContext->bTerminate) {
      pContext->signalled++;

      /* Get the pending FPGA interrupts */
      ADMXRC3_Read(pContext->hCard,	0, 0, REGADDR_ITEST_STATUS, sizeof(istat), &istat);

      /* Clear the pending FPGA interrupts; also rearms the 'finti_l' line by forcing it high for a few cycles */
      ADMXRC3_Write(pContext->hCard, 0, 0, REGADDR_ITEST_CLEAR, sizeof(istat), &istat);
    }
	}

	return 0;
}

static DWORD WINAPI
interruptThreadUber(
  void* pArgument)
{
  ThreadContext* pContext = (ThreadContext*)pArgument;
	DWORD result;
	DWORD istat;
	DWORD dummy = 0; /* value doesn't matter, but avoid compiler whine */

	_tprintf(_T("Interrupt thread (UBER) started\n"));
	
	while (!pContext->bTerminate) {
		result = WaitForSingleObject(pContext->hEvent, INFINITE);
		if (result != WAIT_OBJECT_0) {
      DWORD lastError = GetLastError();

			_tprintf(_T("*** Interrupt thread failed to wait for event (Win32 error %lu(0x%lX))\n"),
        (unsigned long)lastError, (unsigned long)lastError);
			break;
		}

		if (!pContext->bTerminate) {
      pContext->signalled++;

      /* Get the pending FPGA interrupts */
      ADMXRC3_Read(pContext->hCard,	0, 0, REGADDR_UBER_ISTAT, sizeof(istat), &istat);

      /* Clear the pending FPGA interrupts */
      ADMXRC3_Write(pContext->hCard, 0, 0, REGADDR_UBER_ISTAT, sizeof(istat), &istat);

      /* Rearm the 'finti_l' line driven by the FPGA by forcing it high for 1 cycle */
      ADMXRC3_Write(pContext->hCard, 0, 0, REGADDR_UBER_IARM, sizeof(dummy), &dummy);
    }
	}

	return 0;
}

int
_tmain(
  int argc,
  TCHAR *argv[])
{
	const double clockFreq = 33.0e6;
	int ret = 0;
  unsigned int index, serial;
	ADMXRC3_STATUS status;
	ADMXRC3_CARD_INFO cardInfo;
	unsigned long numInterruptGenerated;
	const TCHAR* pFilename;
  Arguments arguments = g_initArguments;
	Option options[3];
	DWORD val32;
  ThreadContext threadContext = {
    ADMXRC3_HANDLE_INVALID_VALUE, /* hCard */
    NULL, /* hEvent */
    FALSE, /* bTerminate */
    0 /* signalled */
  };
  BOOLEAN bEventRegistered = FALSE;
  HANDLE hInterruptThread = NULL;
  DWORD interruptThreadId;
  BOOLEAN bInterruptThreadRunning = FALSE;
  bool_t bUseUber, bShouldConfigure;
  int ch;

  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("uber");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = _T("use uber design instead of itest");
	arguments.nOption = 3;
	arguments.option = options;
	arguments.nParamType = 0;
	arguments.minNParam = 0;

  if (sampleParseCommandLine(_T("itest"), argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bUseUber = options[2].value.booleanVal;

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &threadContext.hCard);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(threadContext.hCard, &cardInfo);
	if (status != ADMXRC3_SUCCESS) {
		_tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
		ret = -1;
		goto done;
	}
	
  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    /* Set local bus clock to nominal frequency */
    status = ADMXRC3_SetClockFrequency(threadContext.hCard, ADMXRC3_CLOCK_LCLK, 0, clockFreq, NULL);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to set LCLK to %.1fMHz: %s\n"), clockFreq / 1.0e6f, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  }

  if (bUseUber) {
    pFilename = sampleGetFpgaFilename(threadContext.hCard, 0, _T("uber"), _T("uber"), &arguments, &bShouldConfigure);
  } else {
    pFilename = sampleGetFpgaFilename(threadContext.hCard, 0, _T("itest"), _T("itest"), &arguments, &bShouldConfigure);
  }
  if (NULL == pFilename) {
    _tprintf(_T("*** Failed to determine bitstream filename.\n"));
    ret = -1;
    goto done;
  }

  if (bShouldConfigure) {
    status = ADMXRC3_ConfigureFromFile(threadContext.hCard, 0, 0, pFilename);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to load the bitstream '%s': %s\n"), pFilename, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  } else {
    /* Running on a model such as the ADM-XRC-6T-ADV8, which doesn't have an FPGA that is configurable via the ADMXRC3 API */
  }

	/* Create an auto-reset event */
	threadContext.hEvent = CreateEvent(NULL, FALSE, FALSE,	NULL);
	if (threadContext.hEvent == NULL) {
		_tprintf(_T("*** Failed to create a Win32 event\n"));
		ret = -1;
		goto done;
	}

	/*
	** Register the event with the card, so that when an FPGA interrupt occurs,
	** the event will be signalled.
	*/
	status = ADMXRC3_RegisterWin32Event(threadContext.hCard, ADMXRC3_EVENT_FPGAINTERRUPT(0), threadContext.hEvent);
	if (status != ADMXRC3_SUCCESS) {
		_tprintf(_T("*** Failed to register Win32 event for FPGA interrupt: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
		ret = -1;
		goto done;
	}
	bEventRegistered = TRUE;

	/* Start up the thread that waits for interrupts. */
	threadContext.signalled = 0;
	threadContext.bTerminate = FALSE;
  hInterruptThread = CreateThread(NULL, 0, bUseUber ? interruptThreadUber : interruptThreadITest, (void*)&threadContext, 0, &interruptThreadId);
	if (hInterruptThread == NULL) {
		_tprintf(_T("*** Failed to start interrupt thread\n"));
		ret = -1;
		goto done;
	}
	bInterruptThreadRunning = TRUE;

  if (bUseUber) {
  	/* Unmask all interrupts in the FPGA (UBER) */
	  val32 = cpu_to_le32(0x0U);
	  status = ADMXRC3_Write(threadContext.hCard, 0, 0, REGADDR_UBER_IMASK, sizeof(val32), &val32);
	  if (status != ADMXRC3_SUCCESS) {
		  _tprintf(_T("*** Write to UBER IMASK register failed: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
		  ret = -1;
		  goto done;
	  }
  } else {
  	/* Enable all interrupts in the FPGA (ITest) */
	  val32 = cpu_to_le32(0xFFFFFFFFU);
	  status = ADMXRC3_Write(threadContext.hCard, 0, 0, REGADDR_ITEST_ENABLE, sizeof(val32), &val32);
	  if (status != ADMXRC3_SUCCESS) {
		  _tprintf(_T("*** Write to ITest ENABLE register failed: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
		  ret = -1;
		  goto done;
	  }
  }

	/*
	** Allow the user to generate interrupts by pressing the enter key.
	** Ctrl-Z or entering 'q' / 'Q' exits this loop.
	*/
	numInterruptGenerated = 0;
	do {
		_tprintf(_T("Enter 'q' to quit, or anything else to generate an interrupt:\n"));
		ch = _gettc(stdin);
		if (EOF == ch || 'q' == tolower(ch)) {
      ch = EOF;
    } else {
      val32 = cpu_to_le32(0x1U);
      status = ADMXRC3_Write(threadContext.hCard, 0, 0, bUseUber ? REGADDR_UBER_TEST : REGADDR_ITEST_TEST, sizeof(val32), &val32);
      if (status != ADMXRC3_SUCCESS) {
        _tprintf(_T("*** Write to FPGA_TEST register failed: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      }

      numInterruptGenerated++;
    }
	} while (EOF != ch);

	_tprintf(_T("Generated %lu interrupts\n"), numInterruptGenerated);
	_tprintf(_T("Interrupt thread saw %lu interrupt(s)\n"), threadContext.signalled);

done:
	if (bInterruptThreadRunning) {
		/* Force the interrupt thread to terminate */
		threadContext.bTerminate = TRUE;
		SetEvent(threadContext.hEvent);
		/* Wait for the interrupt thread to terminate */
		WaitForSingleObject(hInterruptThread, 1000);
	}

	if (threadContext.hEvent != NULL) {
		if (bEventRegistered) {
			/*
			** Unregister the event, so that it will no longer be signalled
			** on an FPGA interrupt.
			*/
			status = ADMXRC3_UnregisterWin32Event(threadContext.hCard, ADMXRC3_EVENT_FPGAINTERRUPT(0), threadContext.hEvent);
			if (status != ADMXRC3_SUCCESS) {
				_tprintf(_T("*** Failed to unregister interrupt event: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
			}
		}
		CloseHandle(threadContext.hEvent);
	}

  if (threadContext.hCard != NULL) {
    ADMXRC3_Close(threadContext.hCard);
    threadContext.hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
