/*
** loader.c - Configure target FPGA with a .bit file and then exit.
**
** (C) Copyright 2012 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _tscanf_s(buffer, ...) _tscanf(buffer, __VA_ARGS__) /* This is OK unless the format string contains %s */
# endif

typedef struct _stat sys_stat_t;

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tmain main
# define _tprintf printf
# define _tcserror strerror
# define _tfopen fopen
# define _tstat stat
# define FALSE (0)
# define TRUE (1)

typedef struct stat sys_stat_t;

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#include <admxrc3.h>

#include <sdk_common.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#if !defined(SIZE_T_MAX)
# define SIZE_T_MAX ((size_t)-1)
#endif

#if !defined(_MSC_VER) || !defined(WIN32)
static int
_tfopen_s(
  FILE** ppFile,
  const TCHAR* pFilename,
  const TCHAR* pMode)
{
  FILE* pFile;

  pFile = _tfopen(pFilename, pMode);
  if (NULL != pFile) {
    *ppFile = pFile;
    return 0;
  } else {
    return errno;
  }
}

static int
_tcserror_s(
  TCHAR* pBuffer,
  size_t n,
  int err)
{
  const TCHAR* p = _tcserror(err);
  size_t i;

  if (n == 0) {
    return 0;
  }
  if (NULL == pBuffer) {
    return EINVAL;
  }
  for (i = 0; i < n; i++) {
    pBuffer[i] = p[i];
    if (p[i] == _T('\0')) {
      break;
    }
  }
  pBuffer[n - 1] = _T('\0');
  return 0;
}
# endif

int
_tmain(
  int argc,
  TCHAR *argv[])
{
  int ret = 0;
  ADMXRC3_STATUS status;
  unsigned int index, serial;
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  Option options[4];
  OptionType paramType[2];
  TCHAR* paramInfo[2];
  TCHAR* paramMeaning[2];
  Arguments arguments = g_initArguments;
  bool_t bIsBinaryFile, bCheckLinkStatus;
  unsigned int targetIndex;
  const TCHAR* pFilename;

  /* Set up expected options */
  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("binary");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = _T("file contains raw SelectMap data (not a .bit file)");
  options[3].key = _T("checklink");
  options[3].type = OptionBoolean;
  options[3].def.booleanVal = TRUE;
  options[3].meaning = _T("check link status");
  paramType[0] = OptionUInt;
  paramInfo[0] = _T("<target FPGA index>");
  paramMeaning[0] = _T("index of the target FPGA to configure");
  paramType[1] = OptionPath;
  paramInfo[1] = _T("<filename.bit>");
  paramMeaning[1] = _T("file with which to configure FPGA");
  arguments.nOption = ARRAY_LENGTH(options);
  arguments.option = options;
  arguments.nParamType = ARRAY_LENGTH(paramType);
  arguments.minNParam = ARRAY_LENGTH(paramType);
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;
  arguments.bNoStdOptions = TRUE;

  if (sampleParseCommandLine(_T("loader"), argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bIsBinaryFile = options[2].value.booleanVal;
  bCheckLinkStatus = options[3].value.booleanVal;
  targetIndex = arguments.param[0].uintVal;
  pFilename = arguments.param[1].pPathVal;

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  if (bIsBinaryFile) {
    FILE* pFile = NULL;
    size_t n, nRead;
    int err;
    TCHAR buf[128];
    void* pData = NULL;
    sys_stat_t s;

    err = _tstat(pFilename, &s);
    if (0 != err) {
      buf[0] = _T('\0');
      _tcserror_s(buf, ARRAY_LENGTH(buf), err);
      _tprintf(_T("*** Failed to stat binary file '%s': %s\n"), pFilename, buf);
      ret = -3;
      goto done;
    }
    if (s.st_size > SIZE_T_MAX) {
      _tprintf(_T("*** Binary file '%s' of %llu(0x%llx) bytes is too large to load\n"), pFilename,
        (unsigned long long)s.st_size, (unsigned long long)s.st_size);
      ret = -4;
      goto done;
    }
    n = (size_t)s.st_size;
    pData = malloc(n);
    if (NULL == pData) {
      _tprintf(_T("*** Failed to allocate buffer of %llu(0x%llx) byte(s) to read binary file '%s'\n"),
        (unsigned long long)n, (unsigned long long)n, pFilename);
      ret = -5;
      goto done;
    }
    err = _tfopen_s(&pFile, pFilename, _T("rb"));
    if (0 != err) {
      buf[0] = _T('\0');
      _tcserror_s(buf, ARRAY_LENGTH(buf), err);
      _tprintf(_T("*** Failed to open binary file '%s' for reading: %s\n"), pFilename, buf);
      free(pData);
      pData = NULL;
      ret = -6;
      goto done;
    }
    nRead = fread(pData, sizeof(uint8_t), n, pFile);
    if (nRead != n) {
      _tprintf(_T("*** Only managed to read %llu(0x%llx) of %llu(0x%llx) bytes for binary file '%s'\n"),
        (unsigned long long)nRead, (unsigned long long)nRead,
        (unsigned long long)n, (unsigned long long)n,
        pFilename);
      fclose(pFile);
      pFile = NULL;
      free(pData);
      pData = NULL;
      ret = -6;
      goto done;
    }
    fclose(pFile);
    pFile = NULL;
    status = ADMXRC3_ConfigureFromBuffer(hCard, targetIndex, bCheckLinkStatus ? 0 : ADMXRC3_CONFIGURE_NOCHECK, pData, n);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to configure target FPGA with binary file '%s': %s\n"), pFilename, ADMXRC3_GetStatusString(status, TRUE));
      free(pData);
      pData = NULL;
      ret = -7;
      goto done;
    }
    free(pData);
    pData = NULL;
  } else {
    status = ADMXRC3_ConfigureFromFile(hCard, targetIndex, bCheckLinkStatus ? 0 : ADMXRC3_CONFIGURE_NOCHECK, pFilename);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to configure target FPGA with '%s': %s\n"), pFilename, ADMXRC3_GetStatusString(status, TRUE));
      ret = -8;
      goto done;
    }
  }

done:
  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
