#pragma once

#include <admxrc3.h>

#include <platform.h>

class ISysMonWnd;
class IDeviceInfoWnd;
class IDeviceStatusWnd;
class ISensorInfoWnd;
class IGraphWnd;
class IGraphWndToolbar;
class IScopeKey;
class IScopeWnd;

class CAdmxrc3Help {
public:
  static const TCHAR* getDatatypeString(ADMXRC3_DATA_TYPE datatype);
  static const TCHAR* getDeviceString(ADMXRC3_FPGA_TYPE device);
  static const TCHAR* getFamilyString(ADMXRC3_FAMILY_TYPE device);
  static const TCHAR* getModelName(ADMXRC3_MODEL_TYPE model);
  static const TCHAR* getSubfamilyString(ADMXRC3_SUBFAMILY_TYPE device);
  static const TCHAR* getUnitString(ADMXRC3_UNIT_TYPE unit);
  static const TCHAR* getWindowDescription(unsigned int index, ADMXRC3_MODEL_TYPE model);
};

struct SensorInfo {
  bool_t m_bValid;
  bool_t m_bKnownDataType;
  ADMXRC3_SENSOR_INFO m_info;
  ADMXRC3_SENSOR_VALUE m_current;
  float m_exponent;
  color_t m_traceColor[2];
};

struct DeviceInfo {
  unsigned int m_numDevice;
  unsigned int m_deviceIndex;
  ADMXRC3_HANDLE m_hDevice;
  ADMXRC3_CARD_INFOEX m_deviceInfo;
  unsigned int m_numSensor;
  unsigned int m_numAvailableSensor;
  SensorInfo* m_pSensors;
};

class IFactory {
public:
  virtual ~IFactory() { /* Nothing to do */ };
  virtual IDeviceInfoWnd* createDeviceInfoWnd(ISysMonWnd* pParent, bool_t bVisible) const = 0;
  virtual IDeviceStatusWnd* createDeviceStatusWnd(ISysMonWnd* pParent, bool_t bVisible) const = 0;
  virtual IGraphWnd* createGraphWnd(ISysMonWnd* pParent, unsigned int index, bool_t bVisible) const = 0;
  virtual IGraphWndToolbar* createGraphWndToolbar(IGraphWnd* pParent, unsigned int index) const = 0;
  virtual IScopeKey* createScopeKey(IGraphWndToolbar* pParent) const = 0;
  virtual IScopeWnd* createScopeWnd(IGraphWnd* pParent) const = 0;
  virtual ISensorInfoWnd* createSensorInfoWnd(ISysMonWnd* pParent, bool_t bVisible) const = 0;
  virtual ISysMonWnd* createSysMonWnd() const = 0;
};

class IControl {
public:
  virtual ~IControl() { /* Nothing to do */ };
  virtual control_handle_t getNativeControl() = 0;
};

class ISysMonWnd {
public:
  virtual ~ISysMonWnd() { /* Nothing to do */ };
  virtual bool_t addGraphWnd(void) = 0;
  virtual bool_t doClearDeviceErrors() = 0;
  virtual bool_t doSelectDeviceIndex(unsigned int deviceIndex) = 0;
  virtual bool_t doSelectTab(unsigned int tabIndex) = 0;
  virtual const DeviceInfo& getDeviceInfo() const = 0;
  virtual IDeviceInfoWnd* getDeviceInfoWnd() const = 0;
  virtual IDeviceStatusWnd* getDeviceStatusWnd() const = 0;
  virtual IGraphWnd* getGraphWnd(unsigned int index) const = 0;
  virtual unsigned int getNumGraphWnd() const = 0;
  virtual unsigned int getUpdatePeriod() const = 0;
  virtual window_handle_t getNativeWindow() = 0;
  virtual ISensorInfoWnd* getSensorInfoWnd() const = 0;
  virtual void idle() = 0;
  virtual bool_t removeGraphWnd(IGraphWnd* pGraphWnd) = 0;
  virtual void requestQuit(void) = 0;
};

class IDeviceInfoWnd : public IControl {
public:
  virtual ~IDeviceInfoWnd() { /* Nothing to do */ };
  virtual void deviceChanged() = 0;
  virtual void numDeviceChanged() = 0;
  virtual ISysMonWnd* getParent() const = 0;
};

class IDeviceStatusWnd : public IControl {
public:
  virtual ~IDeviceStatusWnd() { /* Nothing to do */ };
  virtual void deviceChanged() = 0;
  virtual bool_t doClearDeviceErrors() = 0;
  virtual void numDeviceChanged() = 0;
  virtual ISysMonWnd* getParent() const = 0;
  virtual void update() = 0;
};

class ISensorInfoWnd : public IControl {
public:
  virtual ~ISensorInfoWnd() { /* Nothing to do */ };
  virtual void deviceChanged() = 0;
  virtual void numDeviceChanged() = 0;
  virtual ISysMonWnd* getParent() const = 0;
  virtual void update() = 0;
};

class IGraphWnd : public IControl {
public:
  virtual ~IGraphWnd() { /* Nothing to do */ };
  virtual ISysMonWnd* getParent() const = 0;
  virtual IGraphWndToolbar* getToolbar(void) const = 0;
  virtual IScopeWnd* getScope() const = 0;
  virtual timestamp_t getCurrentTimestamp() const = 0;
  virtual control_handle_t getNativeControl() = 0;
  virtual timestamp_t getStartTimestamp() const = 0;
  virtual void deviceChanged() = 0;
  virtual void update(timestamp_t timestamp) = 0;
  virtual void started() = 0;
  virtual void setIndex(unsigned int index) = 0;
  virtual void viewChanged(timestamp_t interval, timestamp_t minShow, timestamp_t maxShow) = 0;
};

class IGraphWndToolbar : public IControl {
public:
  virtual ~IGraphWndToolbar() { /* Nothing to do */ };
  virtual IGraphWnd* getParent() const = 0;
  virtual bool_t getPaused() const = 0;
  virtual void deviceChanged() = 0;
  virtual void keyClosed() = 0;
  virtual void numGraphWndChanged() = 0;
  virtual void numTraceChanged() = 0;
  virtual void setIndex(unsigned int index) = 0;
  virtual void tracesChanged() = 0;
};

class IScopeKey {
public:
  virtual ~IScopeKey() { /* Nothing to do */ };
  virtual window_handle_t getNativeWindow() = 0;
  virtual void tracesChanged() = 0;
};

class IScopeWnd : public IControl {
public:
  virtual ~IScopeWnd() { /* Nothing to do */ };
  virtual void addTrace(unsigned int sensorIndex, int at = -1) = 0;
  virtual void addTraceDone() = 0;
  virtual void adjustScale(float value) = 0;
  virtual void deviceChanged() = 0;
  virtual void enableTrace(unsigned int traceIndex, bool_t bEnable) = 0;
  virtual unsigned int getNumTrace() = 0;
  virtual IGraphWnd* getParent(void) const = 0;
  virtual unsigned int getTrace(unsigned int traceIndex) = 0;
  virtual void idle() = 0;
  virtual void removeAllTraces() = 0;
  virtual void update(timestamp_t timestamp) = 0;
  virtual void updatePeriodChanged() = 0;
};

extern const IFactory& getFactory();
