#pragma once

#include <math.h>
#include "platform.h"

#define CSTRIPEDLINE_POINTBUFFERSIZE (256) /* This MUST be an even number >= 2 */
#define CSTRIPEDLINE_MAXNUMINTERVAL  (4)

class CStripedLine
{
protected:
  struct Point {
    float m_x, m_y, m_i;
  };
  struct Point2D {
    float m_x, m_y;
  };

  void
  init()
  {
    unsigned int i;

    m_numInterval = 0;
    for (i = 0; i < CSTRIPEDLINE_MAXNUMINTERVAL; i++) {
      m_pPointBuffer[i] = NULL;
      m_pos[i] = 0;
    }

    for (i = 0; i < CSTRIPEDLINE_MAXNUMINTERVAL; i++) {
      m_pPointBuffer[i] = new Point2D[CSTRIPEDLINE_POINTBUFFERSIZE];
    }
  }

public:
  CStripedLine()
  {
    init();
  }
  
  CStripedLine(
    Cairo::RefPtr<Cairo::Context> cr)
  {
    init();
    setContext(cr);
  }
  
  CStripedLine(
    Cairo::RefPtr<Cairo::Context> cr,
    unsigned int n,
    const color_t* pColors)
  {
    init();
    setContext(cr);
    setColors(n, pColors);
  }
  
  CStripedLine(
    unsigned int n,
    const color_t* pColors)
  {
    init();
    setColors(n, pColors);
  }
  
  ~CStripedLine()
  {
    unsigned int i;
    
    for (i = 0; i < CSTRIPEDLINE_MAXNUMINTERVAL; i++) {
      if (NULL != m_pPointBuffer[i]) { 
        delete[] m_pPointBuffer[i];
      }
    }
  }
  
  void
  setContext(Cairo::RefPtr<Cairo::Context> cr)
  {
    m_cairoContext = cr;
  }
  
  void
  setColors(
    unsigned int n,
    const color_t* pColors)
  {
    unsigned int i;

    if (n > CSTRIPEDLINE_MAXNUMINTERVAL) {
      n = CSTRIPEDLINE_MAXNUMINTERVAL;
    }

    m_numInterval = n;
    for (i = 0; i < n; i++) {
      m_colors[i] = pColors[i];
    }
  }
  
  void
  start(float x, float y, float i)
  {
    int fi = (int)floor(i);

    m_last.m_point.m_x = x;
    m_last.m_point.m_y = y;
    m_last.m_point.m_i = i;
    if (fi < 0) {
      m_last.m_which = -fi % m_numInterval;
    } else {
      m_last.m_which = fi % m_numInterval;
    }
    m_next = fi + 1;
  }

  void
  add(float x, float y, float i)
  {
    float dx, dy, di, f, tx, ty;
    Point2D* pPointBuf;
    unsigned int which, pos;
    Point last;

    last = m_last.m_point;
    which = m_last.m_which;
    pPointBuf = m_pPointBuffer[which];

    /* Draw up to start of next interval, if necessary */
    while ((float)m_next < i) {
      pos = m_pos[which];
      /* First point of line segment */
      pPointBuf[pos].m_x = last.m_x;
      pPointBuf[pos].m_y = last.m_y;
      pos++;
      /* Second point of line segment */
      dx = x - last.m_x;
      dy = y - last.m_y;
      di = i - last.m_i;
      f = ((float)m_next - last.m_i) / di;
      tx = last.m_x + f * dx;
      ty = last.m_y + f * dy;
      pPointBuf[pos].m_x = tx;
      pPointBuf[pos].m_y = ty;
      pos++;
      /* Flush out the polylines, if necessary */
      if (pos >= CSTRIPEDLINE_POINTBUFFERSIZE) {
        flush(which, pos);
        pos = 0;
      }
      /* Change to next point buffer */
      m_pos[which++] = pos;
      if (which >= m_numInterval) {
        which = 0;
      }
      pPointBuf = m_pPointBuffer[which];
      /* Update last point */
      last.m_x = tx;
      last.m_y = ty;
      last.m_i = (float)m_next;
      m_next = m_next + 1;
    }

    /* Now draw to requested point */
    pos = m_pos[which];
    /* First point of line segment */
    pPointBuf[pos].m_x = last.m_x;
    pPointBuf[pos].m_y = last.m_y;
    pos++;
    /* Second point of line segment */
    pPointBuf[pos].m_x = x;
    pPointBuf[pos].m_y = y;
    pos++;
    /* Flush out the polylines, if necessary */
    if (pos >= CSTRIPEDLINE_POINTBUFFERSIZE) {
      flush(which, pos);
      pos = 0;
    }
    /* Update last point */
    last.m_x = x;
    last.m_y = y;
    last.m_i = i;

    m_pos[which] = pos;
    m_last.m_which = which;
    m_last.m_point = last;
  }

  void
  finish()
  {
    unsigned int i, pos;

    for (i = 0; i < m_numInterval; i++) {
      pos = m_pos[i];
      if (pos >= 2) {
        flush(i, pos);
      }
      m_pos[i] = 0;
    }
  }

protected:
  void
  flush(
    unsigned int which,
    unsigned int pos)
  {
    color_t color = m_colors[which];
    Point2D* p = m_pPointBuffer[which];
    unsigned int i;

    if (color != TRANSPARENT_COLOR) {
      m_cairoContext->set_source_rgb((double)RED(color) / 255.0, (double)GREEN(color) / 255.0, (double)BLUE(color) / 255.0);
      for (i = 0; i < pos; i += 2) {
        m_cairoContext->move_to(p->m_x, p->m_y);
        p++;
        m_cairoContext->line_to(p->m_x, p->m_y);
        p++;
      }
      m_cairoContext->stroke();
    }
  }
  
protected:
  unsigned int m_numInterval;
  Cairo::RefPtr<Cairo::Context> m_cairoContext;
  color_t m_colors[CSTRIPEDLINE_MAXNUMINTERVAL];
  Point2D* m_pPointBuffer[CSTRIPEDLINE_MAXNUMINTERVAL];
  unsigned int m_pos[CSTRIPEDLINE_MAXNUMINTERVAL];
  struct {
    unsigned int m_which;
    Point m_point;
  } m_last;
  int m_next;
};
