#pragma once

#include <math.h>
#include "platform.h"

#ifdef _DEBUG
# define new DEBUG_NEW
#endif

#define CSTRIPEDLINE_POINTBUFFERSIZE (256) /* This MUST be an even number >= 2 */
#define CSTRIPEDLINE_MAXNUMINTERVAL  (4)

class CStripedLine
{
protected:
  struct Point {
    float m_x, m_y, m_i;
  };

  void
  init()
  {
    unsigned int i;

    m_hDC = NULL;
    m_numInterval = 0;
    for (i = 0; i < CSTRIPEDLINE_MAXNUMINTERVAL; i++) {
      m_hPen[i] = NULL;
      m_pPointBuffer[i] = NULL;
      m_pos[i] = 0;
    }
    m_pCountBuffer = NULL;

    for (i = 0; i < CSTRIPEDLINE_MAXNUMINTERVAL; i++) {
      m_pPointBuffer[i] = new POINT[CSTRIPEDLINE_POINTBUFFERSIZE];
    }
    m_pCountBuffer = new DWORD[CSTRIPEDLINE_POINTBUFFERSIZE >> 1];
    for (i = 0; i < (CSTRIPEDLINE_POINTBUFFERSIZE >> 1); i++) {
      m_pCountBuffer[i] = 2U;
    }
  }

public:
  CStripedLine()
  {
    init();
  }
  
  CStripedLine(
    HDC hDC)
  {
    init();
    setDC(hDC);
  }
  
  CStripedLine(
    HDC hDC,
    unsigned int n,
    const color_t* pColors)
  {
    init();
    setDC(hDC);
    setColors(n, pColors);
  }
  
  CStripedLine(
    unsigned int n,
    const color_t* pColors)
  {
    init();
    setColors(n, pColors);
  }
  
  ~CStripedLine()
  {
    unsigned int i;
    
    for (i = 0; i < CSTRIPEDLINE_MAXNUMINTERVAL; i++) {
      if (NULL != m_pPointBuffer[i]) { 
        delete[] m_pPointBuffer[i];
      }
      if (NULL != m_hPen[i]) {
        DeleteObject(m_hPen[i]);
      }
    }
    if (NULL != m_pCountBuffer) { 
      delete[] m_pCountBuffer;
    }
  }
  
  void
  setDC(HDC hDC)
  {
    m_hDC = hDC;
  }
  
  void
  setColors(
    unsigned int n,
    const color_t* pColors)
  {
    LOGPEN lp;
    unsigned int i;

    if (n > CSTRIPEDLINE_MAXNUMINTERVAL) {
      n = CSTRIPEDLINE_MAXNUMINTERVAL;
    }

    m_numInterval = n;
    for (i = 0; i < n; i++) {
      if (NULL != m_hPen[i]) {
        DeleteObject(m_hPen[i]);
      }
      if (pColors[i] == TRANSPARENT_COLOR) {
        m_hPen[i] = NULL;
      } else {
        ZeroMemory(&lp, sizeof(lp));
        lp.lopnColor = pColors[i];
        lp.lopnStyle = PS_SOLID;
        lp.lopnWidth.x = 1;
        m_hPen[i] = CreatePenIndirect(&lp);
      }
    }
  }
  
  void
  start(float x, float y, float i)
  {
    int fi = (int)floor(i);

    m_last.m_point.m_x = x;
    m_last.m_point.m_y = y;
    m_last.m_point.m_i = i;
    if (fi < 0) {
      m_last.m_which = -fi % m_numInterval;
    } else {
      m_last.m_which = fi % m_numInterval;
    }
    m_next = fi + 1;
  }

  void
  add(float x, float y, float i)
  {
    float dx, dy, di, f, tx, ty;
    POINT ptmp;
    POINT* pPointBuf;
    unsigned int which, pos;
    Point last;

    last = m_last.m_point;
    which = m_last.m_which;
    pPointBuf = m_pPointBuffer[which];

    /* Draw up to start of next interval, if necessary */
    while ((float)m_next < i) {
      pos = m_pos[which];
      /* First point of line segment */
      ptmp.x = (LONG)last.m_x;
      ptmp.y = (LONG)last.m_y;
      pPointBuf[pos++] = ptmp;
      /* Second point of line segment */
      dx = x - last.m_x;
      dy = y - last.m_y;
      di = i - last.m_i;
      f = ((float)m_next - last.m_i) / di;
      tx = last.m_x + f * dx;
      ty = last.m_y + f * dy;
      ptmp.x = (LONG)tx;
      ptmp.y = (LONG)ty;
      pPointBuf[pos++] = ptmp;
      /* Flush out the polylines, if necessary */
      if (pos >= CSTRIPEDLINE_POINTBUFFERSIZE) {
        flush(which, pos);
        pos = 0;
      }
      /* Change to next point buffer */
      m_pos[which++] = pos;
      if (which >= m_numInterval) {
        which = 0;
      }
      pPointBuf = m_pPointBuffer[which];
      /* Update last point */
      last.m_x = tx;
      last.m_y = ty;
      last.m_i = (float)m_next;
      m_next = m_next + 1;
    }

    /* Now draw to requested point */
    pos = m_pos[which];
    /* First point of line segment */
    ptmp.x = (LONG)last.m_x;
    ptmp.y = (LONG)last.m_y;
    pPointBuf[pos++] = ptmp;
    /* Second point of line segment */
    ptmp.x = (LONG)x;
    ptmp.y = (LONG)y;
    pPointBuf[pos++] = ptmp;
    /* Flush out the polylines, if necessary */
    if (pos >= CSTRIPEDLINE_POINTBUFFERSIZE) {
      flush(which, pos);
      pos = 0;
    }
    /* Update last point */
    last.m_x = x;
    last.m_y = y;
    last.m_i = i;

    m_pos[which] = pos;
    m_last.m_which = which;
    m_last.m_point = last;
  }

  void
  finish()
  {
    unsigned int i, pos;

    for (i = 0; i < m_numInterval; i++) {
      pos = m_pos[i];
      if (pos >= 2) {
        flush(i, pos);
      }
      m_pos[i] = 0;
    }
  }

protected:
  void
  flush(
    unsigned int which,
    unsigned int pos)
  {
    HPEN hPen = m_hPen[which];

    if (hPen != NULL) {
      SelectObject(m_hDC, hPen);
      PolyPolyline(m_hDC, m_pPointBuffer[which], m_pCountBuffer, pos >> 1);
    }
  }
  
protected:
  unsigned int m_numInterval;
  HDC m_hDC;
  HPEN m_hPen[CSTRIPEDLINE_MAXNUMINTERVAL];
  POINT* m_pPointBuffer[CSTRIPEDLINE_MAXNUMINTERVAL];
  DWORD* m_pCountBuffer;
  unsigned int m_pos[CSTRIPEDLINE_MAXNUMINTERVAL];
  struct {
    unsigned int m_which;
    Point m_point;
  } m_last;
  int m_next;
};
