/*
** itest.c - example demonstrating interrupt wait functionality in
**           the ADM-XRC API via VxWorks semaphores.
**
** (C) Copyright 2010-2013 Alpha Data
*/

#include <vxWorks.h>
#include <taskLib.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <semaphore.h>

#include <admxrc3.h>

#include "sdk_common.h"
#include "platform_endian.h"

#if (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 9)
# define HAS_VX_USR_ARG_T (1)
# define HAS_TASK_ID (1)
#else
# define HAS_VX_USR_ARG_T (0)
# define HAS_TASK_ID (0)
#endif

#ifndef FALSE
# define FALSE (0)
#endif
#ifndef TRUE
# define TRUE (1)
#endif

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

/*
** FPGA register addresses for the ITest design, in OCP Direct Slave address
** space.
*/
#define REGADDR_ITEST_TEST   (0x000U) /* Test register; write 1 to a bit to generate an interrupt */
#define REGADDR_ITEST_ENABLE (0x180U) /* Enable register in interrupt controller */
#define REGADDR_ITEST_CLEAR  (0x184U) /* Write 1 to a bit to clear corresponding STATUS bit */
#define REGADDR_ITEST_STATUS (0x18CU) /* Status register in interrupt controller */

/*
** FPGA register addresses for the UBER design, in OCP Direct Slave address
** space.
*/
#define REGADDR_UBER_TEST    (0xC0U) /* Test register, write 1 to a bit to generate an interrupt */
#define REGADDR_UBER_ISTAT   (0xC4U) /* Status register */
#define REGADDR_UBER_IMASK   (0xC8U) /* Mask register */
#define REGADDR_UBER_IARM    (0xCCU) /* Rearm register */

typedef struct _ThreadContext {
  ADMXRC3_HANDLE hCard;
  SEM_ID semId;
  BOOL bTerminate;
  unsigned long signalled;
} ThreadContext;

static int
interruptThreadITest(
#if HAS_VX_USR_ARG_T
  _Vx_usr_arg_t arg0,
  _Vx_usr_arg_t arg1,
  _Vx_usr_arg_t arg2,
  _Vx_usr_arg_t arg3,
  _Vx_usr_arg_t arg4,
  _Vx_usr_arg_t arg5,
  _Vx_usr_arg_t arg6,
  _Vx_usr_arg_t arg7,
  _Vx_usr_arg_t arg8,
  _Vx_usr_arg_t arg9)
#else
  int arg0,
  int arg1,
  int arg2,
  int arg3,
  int arg4,
  int arg5,
  int arg6,
  int arg7,
  int arg8,
  int arg9)
#endif
{
  ThreadContext* pContext = (ThreadContext*)arg0;
  ADMXRC3_STATUS status;
  STATUS result;
  uint32_t istat;

  printf("Interrupt thread (ITest) started\n");

  taskSafe();

  while (!pContext->bTerminate) {
    result = semTake(pContext->semId, WAIT_FOREVER);
    if (result != OK) {
      printf("*** Interrupt thread failed to wait for semaphore (errno %d(0x%x))\n", errno, errno);
      break;
    }

    if (!pContext->bTerminate) {
      pContext->signalled++;

      /* Get the pending FPGA interrupts */
      status = ADMXRC3_Read(pContext->hCard, 0, 0, REGADDR_ITEST_STATUS, sizeof(istat), &istat);
      if (ADMXRC3_SUCCESS != status) {
        printf("*** Interrupt thread failed to read ISTAT register: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }

      /* Clear the pending FPGA interrupts; also rearms the 'finti_l' line by forcing it high for a few cycles */
      status = ADMXRC3_Write(pContext->hCard, 0, 0, REGADDR_ITEST_CLEAR, sizeof(istat), &istat);
      if (ADMXRC3_SUCCESS != status) {
        printf("*** Interrupt thread failed to write ISTAT register: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }
    }
  }

  taskUnsafe();

  return 0;
}

static int
interruptThreadUber(
#if HAS_VX_USR_ARG_T
  _Vx_usr_arg_t arg0,
  _Vx_usr_arg_t arg1,
  _Vx_usr_arg_t arg2,
  _Vx_usr_arg_t arg3,
  _Vx_usr_arg_t arg4,
  _Vx_usr_arg_t arg5,
  _Vx_usr_arg_t arg6,
  _Vx_usr_arg_t arg7,
  _Vx_usr_arg_t arg8,
  _Vx_usr_arg_t arg9)
#else
  int arg0,
  int arg1,
  int arg2,
  int arg3,
  int arg4,
  int arg5,
  int arg6,
  int arg7,
  int arg8,
  int arg9)
#endif
{
  ThreadContext* pContext = (ThreadContext*)arg0;
  ADMXRC3_STATUS status;
  STATUS result;
  uint32_t istat;
  uint32_t dummy = 0; /* value doesn't matter, but avoid compiler whine */

  printf("Interrupt thread (UBER) started\n");

  taskSafe();

  while (!pContext->bTerminate) {
    result = semTake(pContext->semId, WAIT_FOREVER);
    if (result != OK) {
      printf("*** Interrupt thread failed to wait for semaphore (errno %d(0x%x))\n", errno, errno);
      break;
    }

    if (!pContext->bTerminate) {
      pContext->signalled++;

      /* Get the pending FPGA interrupts */
      status = ADMXRC3_Read(pContext->hCard, 0, 0, REGADDR_UBER_ISTAT, sizeof(istat), &istat);
      if (ADMXRC3_SUCCESS != status) {
        printf("*** Interrupt thread failed to read ISTAT register: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }

      /* Clear the pending FPGA interrupts */
      status = ADMXRC3_Write(pContext->hCard, 0, 0, REGADDR_UBER_ISTAT, sizeof(istat), &istat);
      if (ADMXRC3_SUCCESS != status) {
        printf("*** Interrupt thread failed to write ISTAT register: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }

      /* Rearm the 'FINTI_L' line driven by the FPGA by forcing it high for 1 cycle */
      status = ADMXRC3_Write(pContext->hCard, 0, 0, REGADDR_UBER_IARM, sizeof(dummy), &dummy);
      if (ADMXRC3_SUCCESS != status) {
        printf("*** Interrupt thread failed to write IARM register: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }
    }
  }

  taskUnsafe();

  return 0;
}

#define FLAG_BYSERIAL (0x1U)  /* Interpret 'index' as a serial number */
#define FLAG_USEUBER  (0x10U) /* Use UBER bitstream instead of ITEST bitstream */

STATUS
admxrc3ITest(
  unsigned int index, /* Index of device to open (OR serial number if 'flags' includes BYSERIAL) */
  unsigned int flags) /* See above flags */
{
  const double clockFreq = 33.0e6;
  STATUS ret = OK;
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFO cardInfo;
  unsigned long numInterruptGenerated;
  char filename[256];
  uint32_t val32;
  ThreadContext threadContext = {
    ADMXRC3_HANDLE_INVALID_VALUE, /* hCard */
    NULL, /* semId */
    FALSE, /* bTerminate */
    0 /* signalled */
  };
  bool_t bOpenBySerial;
  bool_t bSemRegistered = FALSE;
#if HAS_TASK_ID
  TASK_ID interruptThreadId = TASK_ID_ERROR;
#else
  int interruptThreadId = ERROR;
#endif
  bool_t bInterruptThreadRunning = FALSE;
  bool_t bUseUber, bShouldConfigure;
  int ourPriority, ch;

  sampleTaskNice();

  bOpenBySerial = (flags & FLAG_BYSERIAL) ? TRUE : FALSE;
  bUseUber = (flags & FLAG_USEUBER) ? TRUE : FALSE;

  status = sampleOpenCard(index, !bOpenBySerial, index, bOpenBySerial, &threadContext.hCard);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to open card with index/serial %lu: %s\n",
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(threadContext.hCard, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to get card info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    /* Set local bus clock to nominal frequency */
    status = ADMXRC3_SetClockFrequency(threadContext.hCard, ADMXRC3_CLOCK_LCLK, 0, clockFreq, NULL);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to set LCLK to %.1fMHz: %s\n", clockFreq / 1.0e6f, ADMXRC3_GetStatusString(status, TRUE));
      ret = ERROR;
      goto done;
    }
  }

  if (bUseUber) {
    status = sampleGetFpgaFilename(threadContext.hCard, 0, "uber", "uber", ARRAY_LENGTH(filename), filename, &bShouldConfigure);
  } else {
    status = sampleGetFpgaFilename(threadContext.hCard, 0, "itest", "itest", ARRAY_LENGTH(filename), filename, &bShouldConfigure);
  }
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to determine bitstream filename: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  if (bShouldConfigure) {
    status = ADMXRC3_ConfigureFromFile(threadContext.hCard, 0, 0, filename);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to load the bitstream '%s': %s\n", filename, ADMXRC3_GetStatusString(status, TRUE));
      ret = ERROR;
      goto done;
    }
  } else{
    /* Running on a model such as the ADM-XRC-6T-ADV8, which doesn't have an FPGA that is configurable via the ADMXRC3 API */
  }
  
  /* Create a binary semaphore */
  threadContext.semId = semBCreate(0, SEM_EMPTY);
  if (threadContext.semId == NULL) {
    printf("*** Failed to create a VxWorks binary semaphore\n");
    ret = ERROR;
    goto done;
  }

  /*
  ** Register the semaphore with the card, so that when an FPGA interrupt occurs,
  ** the semaphore will be given.
  */
  status = ADMXRC3_RegisterVxwSem(threadContext.hCard, ADMXRC3_EVENT_FPGAINTERRUPT(0), threadContext.semId);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to register Win32 event for FPGA interrupt: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }
  bSemRegistered = TRUE;

  /* Start the thread that waits for interrupts. */
  threadContext.signalled = 0;
  threadContext.bTerminate = FALSE;
  taskPriorityGet(taskIdSelf(), &ourPriority);
#if HAS_VX_USR_ARG_T
  interruptThreadId = taskSpawn("admxrc3InterruptThread", ourPriority - 1, 0, 0x1000, bUseUber ? interruptThreadUber : interruptThreadITest,
    (_Vx_usr_arg_t)&threadContext, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0,
    (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0);
#else
  interruptThreadId = taskSpawn("admxrc3InterruptThread", ourPriority - 1, 0, 0x1000, bUseUber ? interruptThreadUber : interruptThreadITest,
    (int)&threadContext, 0, 0, 0, 0, 0, 0, 0, 0, 0);
#endif
#if HAS_TASK_ID
  if (interruptThreadId == TASK_ID_ERROR) {
#else
  if (interruptThreadId == ERROR) {
#endif
    printf("*** Failed to start interrupt thread\n");
    ret = ERROR;
    goto done;
  }
  bInterruptThreadRunning = TRUE;

  if (bUseUber) {
    /* Unmask all interrupts in the FPGA (UBER) */
    val32 = cpu_to_le32(0x0U);
    status = ADMXRC3_Write(threadContext.hCard, 0, 0, REGADDR_UBER_IMASK, sizeof(val32), &val32);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Write to UBER IMASK register failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  } else {
    /* Enable all interrupts in the FPGA (ITest) */
    val32 = cpu_to_le32(0xFFFFFFFFU);
    status = ADMXRC3_Write(threadContext.hCard, 0, 0, REGADDR_ITEST_ENABLE, sizeof(val32), &val32);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Write to ITest ENABLE register failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  }

  /*
  ** Allow the user to generate interrupts by pressing the enter key.
  ** Ctrl-Z or entering 'q' / 'Q' exits this loop.
  */
  numInterruptGenerated = 0;
  do {
    printf("Enter 'q' to quit, or anything else to generate an interrupt:\n");
    ch = getchar();
    if (EOF == ch || 'q' == tolower(ch)) {
      ch = EOF;
    } else {
      val32 = cpu_to_le32(0x1U);
      status = ADMXRC3_Write(threadContext.hCard, 0, 0, bUseUber ? REGADDR_UBER_TEST : REGADDR_ITEST_TEST, sizeof(val32), &val32);
      if (status != ADMXRC3_SUCCESS) {
        printf("*** Write to ISET register failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }

      numInterruptGenerated++;
    }
  } while (EOF != ch);

  printf("Generated %lu interrupts\n", numInterruptGenerated);
  printf("Interrupt thread saw %lu interrupt(s)\n", threadContext.signalled);

done:
  if (bInterruptThreadRunning) {
    /* Force the interrupt thread to terminate */
    threadContext.bTerminate = TRUE;
    semGive(threadContext.semId);
    /* Wait for the interrupt thread to terminate */
    taskDelete(interruptThreadId);
#if HAS_TASK_ID
    interruptThreadId = TASK_ID_ERROR;
#else
    interruptThreadId = ERROR;
#endif
  }

  if (threadContext.semId != NULL) {
    if (bSemRegistered) {
      /*
      ** Unregister the semaphore, so that it will no longer be signalled
      ** on an FPGA interrupt.
      */
      status = ADMXRC3_UnregisterVxwSem(threadContext.hCard, ADMXRC3_EVENT_FPGAINTERRUPT(0), threadContext.semId);
      if (status != ADMXRC3_SUCCESS) {
        printf("*** Failed to unregister interrupt semaphore: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }
    }
    semDelete(threadContext.semId);
    threadContext.semId = NULL;
  }

  if (threadContext.hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(threadContext.hCard);
    threadContext.hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  return ret;
}
