/*
** loader.c - Configure target FPGA with a .bit file and then exit.
**
** (C) Copyright 2012-2013 Alpha Data
*/

#include <vxWorks.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#include <admxrc3.h>
#include "sdk_common.h"

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#if !defined(SIZE_T_MAX)
# define SIZE_T_MAX ((size_t)-1)
#endif

#define FLAG_BYSERIAL    (0x1U)  /* Interpret 'index' as a serial number */
#define FLAG_BINARY      (0x10U) /* Treat pFilename as binary SelectMap data, not a .bit file */
#define FLAG_NOLINKCHECK (0x20U) /* Don't check link status */

STATUS
admxrc3Loader(
  unsigned int index,       /* Index of device to open (OR serial number if 'flags' includes BYSERIAL) */
  unsigned int flags,       /* See above flags */
  unsigned int targetIndex, /* Index of target FPGA within device */
  char*        pFilename)   /* Name of file with which to configure the FPGA */
{
  int ret = 0;
  ADMXRC3_STATUS status;
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  bool_t bOpenBySerial, bIsBinaryFile, bDontCheckLinkStatus;

  sampleTaskNice();

  bOpenBySerial = (flags & FLAG_BYSERIAL) ? TRUE : FALSE;
  bIsBinaryFile = (flags & FLAG_BINARY) ? TRUE : FALSE;
  bDontCheckLinkStatus = (flags & FLAG_NOLINKCHECK) ? TRUE : FALSE;

  status = sampleOpenCard(index, !bOpenBySerial, index, bOpenBySerial, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to open card with index/serial %lu: %s\n",
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  if (bIsBinaryFile) {
    FILE* pFile = NULL;
    size_t n, nRead;
    int err;
    void* pData = NULL;
    struct stat s;

    err = stat(pFilename, &s);
    if (0 != err) {
      printf("*** Failed to stat binary file '%s': %s\n", pFilename, strerror(errno));
      ret = -3;
      goto done;
    }
    if (s.st_size > SIZE_T_MAX) {
      printf("*** Binary file '%s' of %llu(0x%llx) bytes is too large to load\n", pFilename,
        (unsigned long long)s.st_size, (unsigned long long)s.st_size);
      ret = -4;
      goto done;
    }
    n = (size_t)s.st_size;
    pData = malloc(n);
    if (NULL == pData) {
      printf("*** Failed to allocate buffer of %llu(0x%llx) byte(s) to read binary file '%s'\n",
        (unsigned long long)n, (unsigned long long)n, pFilename);
      ret = -5;
      goto done;
    }
    pFile = fopen(pFilename, "rb");
    if (NULL != pFile) {
      printf("*** Failed to open binary file '%s' for reading: %s\n", pFilename, strerror(errno));
      free(pData);
      pData = NULL;
      ret = -6;
      goto done;
    }
    nRead = fread(pData, sizeof(uint8_t), n, pFile);
    if (nRead != n) {
      printf("*** Only managed to read %llu(0x%llx) of %llu(0x%llx) bytes for binary file '%s'\n",
        (unsigned long long)nRead, (unsigned long long)nRead,
        (unsigned long long)n, (unsigned long long)n,
        pFilename);
      fclose(pFile);
      pFile = NULL;
      free(pData);
      pData = NULL;
      ret = -6;
      goto done;
    }
    fclose(pFile);
    pFile = NULL;
    status = ADMXRC3_ConfigureFromBuffer(hCard, targetIndex, bDontCheckLinkStatus ? ADMXRC3_CONFIGURE_NOCHECK : 0, pData, n);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to configure target FPGA with binary file '%s': %s\n", pFilename, ADMXRC3_GetStatusString(status, TRUE));
      free(pData);
      pData = NULL;
      ret = -7;
      goto done;
    }
    free(pData);
    pData = NULL;
  } else {
    status = ADMXRC3_ConfigureFromFile(hCard, targetIndex, bDontCheckLinkStatus ? ADMXRC3_CONFIGURE_NOCHECK : 0, pFilename);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to configure target FPGA with '%s': %s\n", pFilename, ADMXRC3_GetStatusString(status, TRUE));
      ret = -8;
      goto done;
    }
  }

done:
  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  return ret;
}
