/*
** simpledma.c - Example that verifies the SIMPLEDMA example FPGA design.
**
** (C) Copyright 2012-2013 Alpha Data
*/

#include <vxWorks.h>
#include <taskLib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include <admxrc3.h>

#include "sdk_common.h"
#include "platform_endian.h"
#include "platform_time.h"

/* Default maximum number of individual verification errors to report */
#define MAX_ERROR_REPORTED 20

/* Buffer size used for DMA transfers */
#define BUFFER_SIZE 0x100000U

/* Number of bytes of data to read and display */
#define READ_LENGTH (0x100U)

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))
#define MIN(a, b) ((a) < (b) ? (a) : (b))

static void
strset_s(
  char* pBuffer,
  size_t max,
  char c,
  size_t n)
{
  if (!max) {
    return;
  }
  max--;
  while (max) {
    *pBuffer++ = c;
    max--;
  }
  *pBuffer++ = '\0';
}

static void
dumpMemory(
  unsigned int indent,
  unsigned int nColumn,
  unsigned int radix,
  bool_t bBigEndian,
  bool_t bAlign,
  uint64_t offset,
  const void* pBuffer,
  size_t n)
{
  static const unsigned int ptrSize = sizeof(void*);
  unsigned int ptrStrLen = 3 + 2 * ptrSize + (ptrSize - 1) / 4;
  uint64_t address, limit, addressCol, m;
  unsigned int addressLo, columnMask;
  uint8_t* p;
  uint8_t* p8;
  char line[257];
  char* pPos;
  uint8_t columnBuf[256];
  uint8_t val8;
  uint16_t val16;
  uint32_t val32;
  uint64_t val64;
  unsigned int i, j, k, chunk;
  unsigned int rowBytes;
  int nwritten = 0;
  size_t remaining;

  assert(ptrSize == 4 || ptrSize == 8);

  if (indent > 128) {
    printf("*** dumpMemory: invalid value '%lu' for indent\n", (unsigned long)indent);
    return;
  }
  /* SIZE_T_MAX should really be used here, but is not very portable */
  if (offset > (size_t)-1) {
    printf("*** dumpMemory: offset value '0x%llX' larger than SIZE_T_MAX\n", (unsigned long long)offset);
    return;
  }
  switch (nColumn) {
  case 1:
  case 2:
  case 4:
  case 8:
  case 16:
  case 32:
    break;
  default:
    printf("*** dumpMemory: invalid value '%lu' for nColumn\n", (unsigned long)nColumn);
    return;
  }
  if (radix != 1 && radix != 2 && radix != 4 && radix != 8) {
    printf("*** dumpMemory: invalid value '%lu' for radix\n", (unsigned long)radix);
    return;
  }

  rowBytes = nColumn * radix;

  /* Make sure everything is aligned */
  columnMask = rowBytes - 1U;
  address = (uint64_t)(uintptr_t)pBuffer;
  limit = address + n;
  if (bAlign) {
    addressCol = offset & ~(uint64_t)columnMask;
    addressLo = ((unsigned int)address) & columnMask;
  } else {
    addressCol = offset;
    addressLo = 0;
  }

  m = limit - address;
  p = (uint8_t*)pBuffer;

  pPos = line;
  line[0] = '\0';
  remaining = ARRAY_LENGTH(line);
  strset_s(pPos, remaining, ' ', indent + 1 + ptrStrLen);
  pPos += MIN(indent + 1 + ptrStrLen, remaining);
  remaining -= MIN(indent + 1 + ptrStrLen, remaining);
  for (i = 0; i < rowBytes; i += radix) {
    switch (radix) {
    case 0x1U:
#if (defined(__VXWORKS__) || defined(__vxworks)) && (_WRS_VXWORKS_MAJOR <= 5)
      /* VxWorks 5.x doesn't have snprintf() */
      nwritten = sprintf(pPos, "%02lx ", (unsigned long)i);
#else
      nwritten = snprintf(pPos, remaining, "%02lx ", (unsigned long)i);
#endif
      break;

    case 0x2U:
#if (defined(__VXWORKS__) || defined(__vxworks)) && (_WRS_VXWORKS_MAJOR <= 5)
      /* VxWorks 5.x doesn't have snprintf() */
      nwritten = sprintf(pPos, "  %02lx ", (unsigned long)i);
#else
      nwritten = snprintf(pPos, remaining, "  %02lx ", (unsigned long)i);
#endif
      break;

    case 0x4U:
#if (defined(__VXWORKS__) || defined(__vxworks)) && (_WRS_VXWORKS_MAJOR <= 5)
      /* VxWorks 5.x doesn't have snprintf() */
      nwritten = sprintf(pPos, "      %02lx ", (unsigned long)i);
#else
      nwritten = snprintf(pPos, remaining, "      %02lx ", (unsigned long)i);
#endif
      break;

    case 0x8U:
#if (defined(__VXWORKS__) || defined(__vxworks)) && (_WRS_VXWORKS_MAJOR <= 5)
      /* VxWorks 5.x doesn't have snprintf() */
      nwritten = sprintf(pPos, "               %02lx ", (unsigned long)i);
#else
      nwritten = snprintf(pPos, remaining, "               %02lx ", (unsigned long)i);
#endif
      break;
    }
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }
  }
  if (remaining) {
    *pPos++ = '\n';
    remaining--;
  }
  if (remaining) {
    *pPos++ = '\0';
    remaining--;
  }
  line[ARRAY_LENGTH(line) - 1] = '\0'; /* Ensure line is nul-terminated */
  printf("%s", line);

  while (m) {
    chunk = rowBytes - addressLo;
    if (chunk > m) {
      chunk = (unsigned int)m;
    }

    memset(columnBuf, 0, rowBytes);
    memcpy(columnBuf + addressLo, p, chunk);

    p8 = (uint8_t*)columnBuf;

    pPos = line;
    line[0] = '\0';
    remaining = ARRAY_LENGTH(line);
    strset_s(pPos, remaining, ' ', indent);
    pPos += MIN(indent, remaining);
    remaining -= MIN(indent, remaining);
#if (defined(__VXWORKS__) || defined(__vxworks)) && (_WRS_VXWORKS_MAJOR <= 5)
    /* VxWorks 5.x doesn't have snprintf() */
    if (ptrSize == 4) {
      nwritten = sprintf(pPos, "0x%08lx: ", (unsigned long)(uint32_t)addressCol);
    } else {
      nwritten = sprintf(pPos, "0x%08lx_%08lx: ", (unsigned long)(uint32_t)(addressCol >> 32U), (unsigned long)(uint32_t)addressCol);
    }
#else
    if (ptrSize == 4) {
      nwritten = snprintf(pPos, remaining, "0x%08lx: ", (unsigned long)(uint32_t)addressCol);
    } else {
      nwritten = snprintf(pPos, remaining, "0x%08lx_%08lx: ", (unsigned long)(uint32_t)(addressCol >> 32U), (unsigned long)(uint32_t)addressCol);
    }
#endif
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }

    j = addressLo + chunk;
    for (i = 0; i < rowBytes; i += radix) {
      switch (radix) {
        case 0x1U:
          val8 = *p8++;
          if (i >= addressLo && i < j) {
#if (defined(__VXWORKS__) || defined(__vxworks)) && (_WRS_VXWORKS_MAJOR <= 5)
            /* VxWorks 5.x doesn't have snprintf() */
            nwritten = sprintf(pPos, "%02lx", (unsigned long)val8);
#else
            nwritten = snprintf(pPos, remaining, "%02lx", (unsigned long)val8);
#endif
            if (nwritten > 0) {
              pPos += nwritten;
              remaining -= nwritten;
            }
          } else {
            if (remaining) {
              *pPos++ = '.';
              remaining--;
            }
            if (remaining) {
              *pPos++ = '.';
              remaining--;
            }
          }
          break;

        case 0x2U:
          if (bBigEndian) {
            val16 = (uint16_t)(((uint16_t)p8[0] << 8U) | ((uint16_t)p8[1] << 0U));
          } else {
            val16 = (uint16_t)(((uint16_t)p8[0] << 0U) | ((uint16_t)p8[1] << 8U));
          }
          p8 += 2;
          for (k = 2; k != 0;) {
            k--;
            if (i + k >= addressLo && i + k < j) {
#if (defined(__VXWORKS__) || defined(__vxworks)) && (_WRS_VXWORKS_MAJOR <= 5)
              /* VxWorks 5.x doesn't have snprintf() */
              nwritten = sprintf(pPos, "%02lx", (unsigned long)(uint8_t)(val16 >> (k << 3)));
#else
              nwritten = snprintf(pPos, remaining, "%02lx", (unsigned long)(uint8_t)(val16 >> (k << 3)));
#endif
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
            } else {
              if (remaining) {
                *pPos++ = '.';
                remaining--;
              }
              if (remaining) {
                *pPos++ = '.';
                remaining--;
              }
            }
            if (k == 3) {
              if (remaining) {
                *pPos++ = '_';
                remaining--;
              }
            }
          }
          break;

        case 0x4U:
          if (bBigEndian) {
            val32 = (((uint32_t)p8[0]) << 24U) | (((uint32_t)p8[1]) << 16U) | (((uint32_t)p8[2]) <<  8U) | (((uint32_t)p8[3]) <<  0U);
          } else {
            val32 = (((uint32_t)p8[0]) <<  0U) | (((uint32_t)p8[1]) <<  8U) | (((uint32_t)p8[2]) << 16U) | (((uint32_t)p8[3]) << 24U);
          }
          p8 += 4;
          for (k = 4; k != 0;) {
            k--;
            if (i + k >= addressLo && i + k < j) {
#if (defined(__VXWORKS__) || defined(__vxworks)) && (_WRS_VXWORKS_MAJOR <= 5)
              /* VxWorks 5.x doesn't have snprintf() */
              nwritten = sprintf(pPos, "%02lx", (unsigned long)(uint8_t)(val32 >> (k << 3)));
#else
              nwritten = snprintf(pPos, remaining, "%02lx", (unsigned long)(uint8_t)(val32 >> (k << 3)));
#endif
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
            } else {
              if (remaining) {
                *pPos++ = '.';
                remaining--;
              }
              if (remaining) {
                *pPos++ = '.';
                remaining--;
              }
            }
          }
          break;

        case 0x8U:
          if (bBigEndian) {
            val64  = (((uint64_t)p8[0]) << 56U) | (((uint64_t)p8[1]) << 48U) | (((uint64_t)p8[2]) << 40U) | (((uint64_t)p8[3]) << 32U);
            val64 |= (((uint64_t)p8[4]) << 24U) | (((uint64_t)p8[5]) << 16U) | (((uint64_t)p8[6]) <<  8U) | (((uint64_t)p8[7]) <<  0U);
          } else {
            val64  = (((uint64_t)p8[0]) <<  0U) | (((uint64_t)p8[1]) <<  8U) | (((uint64_t)p8[2]) << 16U) | (((uint64_t)p8[3]) << 24U);
            val64 |= (((uint64_t)p8[4]) << 32U) | (((uint64_t)p8[5]) << 40U) | (((uint64_t)p8[6]) << 48U) | (((uint64_t)p8[7]) << 56U);
          }
          p8 += 8;
          for (k = 8; k != 0;) {
            k--;
            if (i + k >= addressLo && i + k < j) {
#if (defined(__VXWORKS__) || defined(__vxworks)) && (_WRS_VXWORKS_MAJOR <= 5)
              /* VxWorks 5.x doesn't have snprintf() */
              nwritten = sprintf(pPos, "%02lx", (unsigned long)(uint8_t)(val64 >> (k << 3)));
#else
              nwritten = snprintf(pPos, remaining, "%02lx", (unsigned long)(uint8_t)(val64 >> (k << 3)));
#endif
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
            } else {
              if (remaining) {
                *pPos++ = '.';
                remaining--;
              }
              if (remaining) {
                *pPos++ = '.';
                remaining--;
              }
            }
            if (k == 4) {
              if (remaining) {
                *pPos++ = '_';
                remaining--;
              }
            }
          }
          break;
      }
      if (remaining) {
        *pPos++ = ' ';
        remaining--;
      }
    }

    p8 = (uint8_t*)columnBuf;
    for (i = 0; i < rowBytes; i++) {
      char c = (char)*p8++;

      if (remaining) {
        if (isprint((int)c) && c != '\t') {
          *pPos++ = c;
        } else {
          *pPos++ = '.';
        }
        remaining--;
      }
    }
    if (remaining) {
      *pPos++ = '\n';
      remaining--;
    }
    if (remaining) {
      *pPos++ = '\0';
      remaining--;
    }
    line[ARRAY_LENGTH(line) - 1] = '\0'; /* Ensure line is nul-terminated */
    printf("%s", line);
    addressLo = 0;
    addressCol += rowBytes;
    m -= chunk;
    p += chunk;
    offset += chunk;
  }
}

#define FLAG_BYSERIAL (0x1U)  /* Interpret 'index' as a serial number */

int
admxrc3SimpleDma(
  unsigned int index,          /* Index of device to open (OR serial number if 'flags' includes BYSERIAL) */
  unsigned int flags,          /* See above flags */
  uint32_t     localAddressLo, /* Low 32 bits of local address at which to being reading */
  uint32_t     localAddressHi) /* High 32 bits of local address at which to being reading */
{
  int ret = 0;
  ADMXRC3_STATUS status;
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_CARD_INFO cardInfo;
  uint8_t* pCheckBuffer = NULL;
  ADMXRC3_BUFFER_HANDLE hCheckBuffer = 0;
  char filename[256];
  unsigned int numError, maxError, offset;
  bool_t bShouldConfigure, bOpenBySerial;
  unsigned int direction;
  uint64_t localAddress;
  uint32_t* p32;
  
  sampleTaskNice();

  bOpenBySerial = (flags & FLAG_BYSERIAL) ? TRUE : FALSE;
  localAddress = (uint64_t)localAddressLo | ((uint64_t)localAddressHi << 32);

  status = sampleOpenCard(index, !bOpenBySerial, index, bOpenBySerial, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to open card with index/serial %lu: %s\n",
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }
  
  status = ADMXRC3_GetCardInfo(hDevice, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to get card info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    /* Set local bus clock to nominal frequency */
    status = ADMXRC3_SetClockFrequency(hDevice, ADMXRC3_CLOCK_LCLK, 0, 33.0e6, NULL);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to set LCLK to %.1f MHz: %s\n", 33.0, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  }

  pCheckBuffer = (uint8_t*)malloc(BUFFER_SIZE);
  if (NULL == pCheckBuffer) {
    printf("*** Failed to allocate check buffer of 0x%x bytes.\n", (unsigned int)BUFFER_SIZE);
    ret = -1;
    goto done;
  }

  status = ADMXRC3_Lock(hDevice, pCheckBuffer, BUFFER_SIZE, &hCheckBuffer);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to lock check buffer for DMA transfers: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  status = sampleGetFpgaFilename(hDevice, 0, "simpledma", "simpledma", ARRAY_LENGTH(filename), filename, &bShouldConfigure);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to determine bitstream filename: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  if (bShouldConfigure) {
    status = ADMXRC3_ConfigureFromFile(hDevice, 0, 0, filename);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to configure target FPGA with '%s': %s\n", filename, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  } else {
    /* Running on a model such as the ADM-XRC-6T-ADV8, which doesn't have an FPGA that is configurable via the ADMXRC3 API */
  }

  printf("Reading using DMA channel 0 at OCP address 0x%llx...\n", (unsigned long long)localAddress);
  status = ADMXRC3_ReadDMAEx(hDevice, 0, 0, pCheckBuffer, READ_LENGTH, localAddress);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to read using DMA channel 0: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }
  printf("Dump of data read from OCP address 0x%llx + 0x%x B:\n", (unsigned long long)localAddress, READ_LENGTH);
  dumpMemory(0, 4, 4, FALSE, FALSE, localAddress, pCheckBuffer, READ_LENGTH);

  /* Check the data */
  p32 = (uint32_t*)pCheckBuffer;
  numError = 0U;
  for (offset = 0U; offset < READ_LENGTH; offset += 4) {
    uint32_t exp32, val32;

    exp32 = offset >> 2;
    val32 = *p32++;
    if (val32 != exp32) {
      if (numError < maxError) {
        printf("*** Error at offset 0x%08x: expected=0x%08lx actual=0x%08lx\n",
          offset, (unsigned long)exp32, (unsigned long)val32);
      } else if (numError == MAX_ERROR_REPORTED) {
        printf("*** More errors...\n");
      }
      numError++;
    }
  }
  if (numError) {
    printf("*** Number of errors in data: %u\n", numError);
  }

  printf("Measuring throughput...                 \n");
  for (direction = 0; direction < 2; direction++) {
    admxrc3_time_t start, finish, now;
    uint64_t total = 0;
    double elapsed, rate;

    start = timeGet();
    finish = timeAdd(start, timeMake(2, 0, FALSE));
    do {
      if (direction == 0) {
        status = ADMXRC3_WriteDMALockedEx(hDevice, 0, 0, hCheckBuffer, 0, BUFFER_SIZE, 0);
        if (ADMXRC3_SUCCESS != status) {
          printf("*** Failed to write memory: %s\n", ADMXRC3_GetStatusString(status, TRUE));
          ret = -1;
          goto done;
        }
      } else {
        status = ADMXRC3_ReadDMALockedEx(hDevice, 0, 0, hCheckBuffer, 0, BUFFER_SIZE, 0);
        if (ADMXRC3_SUCCESS != status) {
          printf("*** Failed to read memory: %s\n", ADMXRC3_GetStatusString(status, TRUE));
          ret = -1;
          goto done;
        }
      }
      total += BUFFER_SIZE;
      now = timeGet();
    } while (timeBefore(now, finish));
    elapsed = timeToDouble(timeSubtract(now, start));
    rate = (double)total / elapsed;
    printf("Throughput %s is %.1f MiB/s\n", (direction == 0) ? "from host to FPGA" : "from FPGA to host", rate / 1048576.0);
  }

  printf("PASSED                                  \n");

done:
  if (numError) {
    printf("*** FAILED with %llu (0x%llx) error(s)\n", (unsigned long long)numError, (unsigned long long)numError);
    ret = -3;
  }

  if (0 != hCheckBuffer) {
    ADMXRC3_Unlock(hDevice, hCheckBuffer);
    hCheckBuffer = 0;
  }
  if (NULL != pCheckBuffer) {
    free(pCheckBuffer);
    pCheckBuffer = NULL;
  }
  
  if (hDevice != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hDevice);
    hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  return ret;
}
